from __future__ import annotations

from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

from s2clientprotocol.spatial_pb2 import ActionSpatial, ObservationFeatureLayer, ObservationRender

from .common_pb2 import AvailableAbility, Point2D, Size2DI
from .data_pb2 import AbilityData, BuffData, EffectData, UnitTypeData, UpgradeData
from .debug_pb2 import DebugCommand
from .query_pb2 import RequestQuery, ResponseQuery
from .raw_pb2 import ActionRaw, ObservationRaw, StartRaw
from .score_pb2 import Score
from .ui_pb2 import ActionUI, ObservationUI

class Request(Message):
    create_game: RequestCreateGame
    join_game: RequestJoinGame
    restart_game: RequestRestartGame
    start_replay: RequestStartReplay
    leave_game: RequestLeaveGame
    quick_save: RequestQuickSave
    quick_load: RequestQuickLoad
    quit: RequestQuit
    game_info: RequestGameInfo
    observation: RequestObservation
    action: RequestAction
    obs_action: RequestObserverAction
    step: RequestStep
    data: RequestData
    query: RequestQuery
    save_replay: RequestSaveReplay
    map_command: RequestMapCommand
    replay_info: RequestReplayInfo
    available_maps: RequestAvailableMaps
    save_map: RequestSaveMap
    ping: RequestPing
    debug: RequestDebug
    id: int
    def __init__(
        self,
        create_game: RequestCreateGame = ...,
        join_game: RequestJoinGame = ...,
        restart_game: RequestRestartGame = ...,
        start_replay: RequestStartReplay = ...,
        leave_game: RequestLeaveGame = ...,
        quick_save: RequestQuickSave = ...,
        quick_load: RequestQuickLoad = ...,
        quit: RequestQuit = ...,
        game_info: RequestGameInfo = ...,
        observation: RequestObservation = ...,
        action: RequestAction = ...,
        obs_action: RequestObserverAction = ...,
        step: RequestStep = ...,
        data: RequestData = ...,
        query: RequestQuery = ...,
        save_replay: RequestSaveReplay = ...,
        map_command: RequestMapCommand = ...,
        replay_info: RequestReplayInfo = ...,
        available_maps: RequestAvailableMaps = ...,
        save_map: RequestSaveMap = ...,
        ping: RequestPing = ...,
        debug: RequestDebug = ...,
        id: int = ...,
    ) -> None: ...

class Response(Message):
    create_game: ResponseCreateGame
    join_game: ResponseJoinGame
    restart_game: ResponseRestartGame
    start_replay: ResponseStartReplay
    leave_game: ResponseLeaveGame
    quick_save: ResponseQuickSave
    quick_load: ResponseQuickLoad
    quit: ResponseQuit
    game_info: ResponseGameInfo
    observation: ResponseObservation
    action: ResponseAction
    obs_action: ResponseObserverAction
    step: ResponseStep
    data: ResponseData
    query: ResponseQuery
    save_replay: ResponseSaveReplay
    replay_info: ResponseReplayInfo
    available_maps: ResponseAvailableMaps
    save_map: ResponseSaveMap
    map_command: ResponseMapCommand
    ping: ResponsePing
    debug: ResponseDebug
    id: int
    error: Iterable[str]
    status: int
    def __init__(
        self,
        create_game: ResponseCreateGame = ...,
        join_game: ResponseJoinGame = ...,
        restart_game: ResponseRestartGame = ...,
        start_replay: ResponseStartReplay = ...,
        leave_game: ResponseLeaveGame = ...,
        quick_save: ResponseQuickSave = ...,
        quick_load: ResponseQuickLoad = ...,
        quit: ResponseQuit = ...,
        game_info: ResponseGameInfo = ...,
        observation: ResponseObservation = ...,
        action: ResponseAction = ...,
        obs_action: ResponseObserverAction = ...,
        step: ResponseStep = ...,
        data: ResponseData = ...,
        query: ResponseQuery = ...,
        save_replay: ResponseSaveReplay = ...,
        replay_info: ResponseReplayInfo = ...,
        available_maps: ResponseAvailableMaps = ...,
        save_map: ResponseSaveMap = ...,
        map_command: ResponseMapCommand = ...,
        ping: ResponsePing = ...,
        debug: ResponseDebug = ...,
        id: int = ...,
        error: Iterable[str] = ...,
        status: int = ...,
    ) -> None: ...

class Status(Enum):
    launched: int
    init_game: int
    in_game: int
    in_replay: int
    ended: int
    quit: int
    unknown: int

class RequestCreateGame(Message):
    local_map: LocalMap
    battlenet_map_name: str
    player_setup: Iterable[PlayerSetup]
    disable_fog: bool
    random_seed: int
    realtime: bool
    def __init__(
        self,
        local_map: LocalMap = ...,
        battlenet_map_name: str = ...,
        player_setup: Iterable[PlayerSetup] = ...,
        disable_fog: bool = ...,
        random_seed: int = ...,
        realtime: bool = ...,
    ) -> None: ...

class LocalMap(Message):
    map_path: str
    map_data: bytes
    def __init__(self, map_path: str = ..., map_data: bytes = ...) -> None: ...

class ResponseCreateGame(Message):
    class Error(Enum):
        MissingMap: int
        InvalidMapPath: int
        InvalidMapData: int
        InvalidMapName: int
        InvalidMapHandle: int
        MissingPlayerSetup: int
        InvalidPlayerSetup: int
        MultiplayerUnsupported: int

    error: int
    error_details: str
    def __init__(self, error: int = ..., error_details: str = ...) -> None: ...

class RequestJoinGame(Message):
    race: int
    observed_player_id: int
    options: InterfaceOptions
    server_ports: PortSet
    client_ports: Iterable[PortSet]
    shared_port: int
    player_name: str
    host_ip: str
    def __init__(
        self,
        race: int = ...,
        observed_player_id: int = ...,
        options: InterfaceOptions = ...,
        server_ports: PortSet = ...,
        client_ports: Iterable[PortSet] = ...,
        shared_port: int = ...,
        player_name: str = ...,
        host_ip: str = ...,
    ) -> None: ...

class PortSet(Message):
    game_port: int
    base_port: int
    def __init__(self, game_port: int = ..., base_port: int = ...) -> None: ...

class ResponseJoinGame(Message):
    class Error(Enum):
        MissingParticipation: int
        InvalidObservedPlayerId: int
        MissingOptions: int
        MissingPorts: int
        GameFull: int
        LaunchError: int
        FeatureUnsupported: int
        NoSpaceForUser: int
        MapDoesNotExist: int
        CannotOpenMap: int
        ChecksumError: int
        NetworkError: int
        OtherError: int

    player_id: int
    error: int
    error_details: str
    def __init__(self, player_id: int = ..., error: int = ..., error_details: str = ...) -> None: ...

class RequestRestartGame(Message):
    def __init__(self) -> None: ...

class ResponseRestartGame(Message):
    class Error(Enum):
        LaunchError: int

    error: int
    error_details: str
    need_hard_reset: bool
    def __init__(self, error: int = ..., error_details: str = ..., need_hard_reset: bool = ...) -> None: ...

class RequestStartReplay(Message):
    replay_path: str
    replay_data: bytes
    map_data: bytes
    observed_player_id: int
    options: InterfaceOptions
    disable_fog: bool
    realtime: bool
    record_replay: bool
    def __init__(
        self,
        replay_path: str = ...,
        replay_data: bytes = ...,
        map_data: bytes = ...,
        observed_player_id: int = ...,
        options: InterfaceOptions = ...,
        disable_fog: bool = ...,
        realtime: bool = ...,
        record_replay: bool = ...,
    ) -> None: ...

class ResponseStartReplay(Message):
    class Error(Enum):
        MissingReplay: int
        InvalidReplayPath: int
        InvalidReplayData: int
        InvalidMapData: int
        InvalidObservedPlayerId: int
        MissingOptions: int
        LaunchError: int

    error: int
    error_details: str
    def __init__(self, error: int = ..., error_details: str = ...) -> None: ...

class RequestMapCommand(Message):
    trigger_cmd: str
    def __init__(self, trigger_cmd: str = ...) -> None: ...

class ResponseMapCommand(Message):
    class Error(Enum):
        NoTriggerError: int

    error: int
    error_details: str
    def __init__(self, error: int = ..., error_details: str = ...) -> None: ...

class RequestLeaveGame(Message):
    def __init__(self) -> None: ...

class ResponseLeaveGame(Message):
    def __init__(self) -> None: ...

class RequestQuickSave(Message):
    def __init__(self) -> None: ...

class ResponseQuickSave(Message):
    def __init__(self) -> None: ...

class RequestQuickLoad(Message):
    def __init__(self) -> None: ...

class ResponseQuickLoad(Message):
    def __init__(self) -> None: ...

class RequestQuit(Message):
    def __init__(self) -> None: ...

class ResponseQuit(Message):
    def __init__(self) -> None: ...

class RequestGameInfo(Message):
    def __init__(self) -> None: ...

class ResponseGameInfo(Message):
    map_name: str
    mod_names: Iterable[str]
    local_map_path: str
    player_info: Iterable[PlayerInfo]
    start_raw: StartRaw
    options: InterfaceOptions
    def __init__(
        self,
        map_name: str = ...,
        mod_names: Iterable[str] = ...,
        local_map_path: str = ...,
        player_info: Iterable[PlayerInfo] = ...,
        start_raw: StartRaw = ...,
        options: InterfaceOptions = ...,
    ) -> None: ...

class RequestObservation(Message):
    disable_fog: bool
    game_loop: int
    def __init__(self, disable_fog: bool = ..., game_loop: int = ...) -> None: ...

class ResponseObservation(Message):
    actions: Iterable[Action]
    action_errors: Iterable[ActionError]
    observation: Observation
    player_result: Iterable[PlayerResult]
    chat: Iterable[ChatReceived]
    def __init__(
        self,
        actions: Iterable[Action] = ...,
        action_errors: Iterable[ActionError] = ...,
        observation: Observation = ...,
        player_result: Iterable[PlayerResult] = ...,
        chat: Iterable[ChatReceived] = ...,
    ) -> None: ...

class ChatReceived(Message):
    player_id: int
    message: str
    def __init__(self, player_id: int = ..., message: str = ...) -> None: ...

class RequestAction(Message):
    actions: Iterable[Action]
    def __init__(self, actions: Iterable[Action] = ...) -> None: ...

class ResponseAction(Message):
    result: Iterable[int]
    def __init__(self, result: Iterable[int] = ...) -> None: ...

class RequestObserverAction(Message):
    actions: Iterable[ObserverAction]
    def __init__(self, actions: Iterable[ObserverAction] = ...) -> None: ...

class ResponseObserverAction(Message):
    def __init__(self) -> None: ...

class RequestStep(Message):
    count: int
    def __init__(self, count: int = ...) -> None: ...

class ResponseStep(Message):
    simulation_loop: int
    def __init__(self, simulation_loop: int = ...) -> None: ...

class RequestData(Message):
    ability_id: bool
    unit_type_id: bool
    upgrade_id: bool
    buff_id: bool
    effect_id: bool
    def __init__(
        self,
        ability_id: bool = ...,
        unit_type_id: bool = ...,
        upgrade_id: bool = ...,
        buff_id: bool = ...,
        effect_id: bool = ...,
    ) -> None: ...

class ResponseData(Message):
    abilities: Iterable[AbilityData]
    units: Iterable[UnitTypeData]
    upgrades: Iterable[UpgradeData]
    buffs: Iterable[BuffData]
    effects: Iterable[EffectData]
    def __init__(
        self,
        abilities: Iterable[AbilityData] = ...,
        units: Iterable[UnitTypeData] = ...,
        upgrades: Iterable[UpgradeData] = ...,
        buffs: Iterable[BuffData] = ...,
        effects: Iterable[EffectData] = ...,
    ) -> None: ...

class RequestSaveReplay(Message):
    def __init__(self) -> None: ...

class ResponseSaveReplay(Message):
    data: bytes
    def __init__(self, data: bytes = ...) -> None: ...

class RequestReplayInfo(Message):
    replay_path: str
    replay_data: bytes
    download_data: bool
    def __init__(
        self,
        replay_path: str = ...,
        replay_data: bytes = ...,
        download_data: bool = ...,
    ) -> None: ...

class PlayerInfoExtra(Message):
    player_info: PlayerInfo
    player_result: PlayerResult
    player_mmr: int
    player_apm: int
    def __init__(
        self,
        player_info: PlayerInfo = ...,
        player_result: PlayerResult = ...,
        player_mmr: int = ...,
        player_apm: int = ...,
    ) -> None: ...

class ResponseReplayInfo(Message):
    class Error(Enum):
        MissingReplay: int
        InvalidReplayPath: int
        InvalidReplayData: int
        ParsingError: int
        DownloadError: int

    map_name: str
    local_map_path: str
    player_info: Iterable[PlayerInfoExtra]
    game_duration_loops: int
    game_duration_seconds: float
    game_version: str
    data_version: str
    data_build: int
    base_build: int
    error: int
    error_details: str
    def __init__(
        self,
        map_name: str = ...,
        local_map_path: str = ...,
        player_info: Iterable[PlayerInfoExtra] = ...,
        game_duration_loops: int = ...,
        game_duration_seconds: float = ...,
        game_version: str = ...,
        data_version: str = ...,
        data_build: int = ...,
        base_build: int = ...,
        error: int = ...,
        error_details: str = ...,
    ) -> None: ...

class RequestAvailableMaps(Message):
    def __init__(self) -> None: ...

class ResponseAvailableMaps(Message):
    local_map_paths: Iterable[str]
    battlenet_map_names: Iterable[str]
    def __init__(self, local_map_paths: Iterable[str] = ..., battlenet_map_names: Iterable[str] = ...) -> None: ...

class RequestSaveMap(Message):
    map_path: str
    map_data: bytes
    def __init__(self, map_path: str = ..., map_data: bytes = ...) -> None: ...

class ResponseSaveMap(Message):
    class Error(Enum):
        InvalidMapData: int

    error: int
    def __init__(self, error: int = ...) -> None: ...

class RequestPing(Message):
    def __init__(self) -> None: ...

class ResponsePing(Message):
    game_version: str
    data_version: str
    data_build: int
    base_build: int
    def __init__(
        self,
        game_version: str = ...,
        data_version: str = ...,
        data_build: int = ...,
        base_build: int = ...,
    ) -> None: ...

class RequestDebug(Message):
    debug: Iterable[DebugCommand]
    def __init__(self, debug: Iterable[DebugCommand] = ...) -> None: ...

class ResponseDebug(Message):
    def __init__(self) -> None: ...

class Difficulty(Enum):
    VeryEasy: int
    Easy: int
    Medium: int
    MediumHard: int
    Hard: int
    Harder: int
    VeryHard: int
    CheatVision: int
    CheatMoney: int
    CheatInsane: int

class PlayerType(Enum):
    Participant: int
    Computer: int
    Observer: int

class AIBuild(Enum):
    RandomBuild: int
    Rush: int
    Timing: int
    Power: int
    Macro: int
    Air: int

class PlayerSetup(Message):
    type: int
    race: int
    difficulty: int
    player_name: str
    ai_build: int
    def __init__(
        self,
        type: int = ...,
        race: int = ...,
        difficulty: int = ...,
        player_name: str = ...,
        ai_build: int = ...,
    ) -> None: ...

class SpatialCameraSetup(Message):
    resolution: Size2DI
    minimap_resolution: Size2DI
    width: float
    crop_to_playable_area: bool
    allow_cheating_layers: bool
    def __init__(
        self,
        resolution: Size2DI = ...,
        minimap_resolution: Size2DI = ...,
        width: float = ...,
        crop_to_playable_area: bool = ...,
        allow_cheating_layers: bool = ...,
    ) -> None: ...

class InterfaceOptions(Message):
    raw: bool
    score: bool
    feature_layer: SpatialCameraSetup
    render: SpatialCameraSetup
    show_cloaked: bool
    show_burrowed_shadows: bool
    show_placeholders: bool
    raw_affects_selection: bool
    raw_crop_to_playable_area: bool
    def __init__(
        self,
        raw: bool = ...,
        score: bool = ...,
        feature_layer: SpatialCameraSetup = ...,
        render: SpatialCameraSetup = ...,
        show_cloaked: bool = ...,
        show_burrowed_shadows: bool = ...,
        show_placeholders: bool = ...,
        raw_affects_selection: bool = ...,
        raw_crop_to_playable_area: bool = ...,
    ) -> None: ...

class PlayerInfo(Message):
    player_id: int
    type: int
    race_requested: int
    race_actual: int
    difficulty: int
    ai_build: int
    player_name: str
    def __init__(
        self,
        player_id: int = ...,
        type: int = ...,
        race_requested: int = ...,
        race_actual: int = ...,
        difficulty: int = ...,
        ai_build: int = ...,
        player_name: str = ...,
    ) -> None: ...

class PlayerCommon(Message):
    player_id: int
    minerals: int
    vespene: int
    food_cap: int
    food_used: int
    food_army: int
    food_workers: int
    idle_worker_count: int
    army_count: int
    warp_gate_count: int
    larva_count: int
    def __init__(
        self,
        player_id: int = ...,
        minerals: int = ...,
        vespene: int = ...,
        food_cap: int = ...,
        food_used: int = ...,
        food_army: int = ...,
        food_workers: int = ...,
        idle_worker_count: int = ...,
        army_count: int = ...,
        warp_gate_count: int = ...,
        larva_count: int = ...,
    ) -> None: ...

class Observation(Message):
    game_loop: int
    player_common: PlayerCommon
    alerts: Iterable[int]
    abilities: Iterable[AvailableAbility]
    score: Score
    raw_data: ObservationRaw
    feature_layer_data: ObservationFeatureLayer
    render_data: ObservationRender
    ui_data: ObservationUI
    def __init__(
        self,
        game_loop: int = ...,
        player_common: PlayerCommon = ...,
        alerts: Iterable[int] = ...,
        abilities: Iterable[AvailableAbility] = ...,
        score: Score = ...,
        raw_data: ObservationRaw = ...,
        feature_layer_data: ObservationFeatureLayer = ...,
        render_data: ObservationRender = ...,
        ui_data: ObservationUI = ...,
    ) -> None: ...

class Action(Message):
    action_raw: ActionRaw
    action_feature_layer: ActionSpatial
    action_render: ActionSpatial
    action_ui: ActionUI
    action_chat: ActionChat
    game_loop: int
    def __init__(
        self,
        action_raw: ActionRaw = ...,
        action_feature_layer: ActionSpatial = ...,
        action_render: ActionSpatial = ...,
        action_ui: ActionUI = ...,
        action_chat: ActionChat = ...,
        game_loop: int = ...,
    ) -> None: ...

class Channel(Enum):
    Broadcast: int
    Team: int

class ActionChat(Message):
    channel: int
    message: str
    def __init__(self, channel: int = ..., message: str = ...) -> None: ...

class ActionError(Message):
    unit_tag: int
    ability_id: int
    result: int
    def __init__(self, unit_tag: int = ..., ability_id: int = ..., result: int = ...) -> None: ...

class ObserverAction(Message):
    player_perspective: ActionObserverPlayerPerspective
    camera_move: ActionObserverCameraMove
    camera_follow_player: ActionObserverCameraFollowPlayer
    camera_follow_units: ActionObserverCameraFollowUnits
    def __init__(
        self,
        player_perspective: ActionObserverPlayerPerspective = ...,
        camera_move: ActionObserverCameraMove = ...,
        camera_follow_player: ActionObserverCameraFollowPlayer = ...,
        camera_follow_units: ActionObserverCameraFollowUnits = ...,
    ) -> None: ...

class ActionObserverPlayerPerspective(Message):
    player_id: int
    def __init__(self, player_id: int = ...) -> None: ...

class ActionObserverCameraMove(Message):
    world_pos: Point2D
    distance: float
    def __init__(self, world_pos: Point2D = ..., distance: float = ...) -> None: ...

class ActionObserverCameraFollowPlayer(Message):
    player_id: int
    def __init__(self, player_id: int = ...) -> None: ...

class ActionObserverCameraFollowUnits(Message):
    unit_tags: Iterable[int]
    def __init__(self, unit_tags: Iterable[int] = ...) -> None: ...

class Alert(Enum):
    AlertError: int
    AddOnComplete: int
    BuildingComplete: int
    BuildingUnderAttack: int
    LarvaHatched: int
    MergeComplete: int
    MineralsExhausted: int
    MorphComplete: int
    MothershipComplete: int
    MULEExpired: int
    NuclearLaunchDetected: int
    NukeComplete: int
    NydusWormDetected: int
    ResearchComplete: int
    TrainError: int
    TrainUnitComplete: int
    TrainWorkerComplete: int
    TransformationComplete: int
    UnitUnderAttack: int
    UpgradeComplete: int
    VespeneExhausted: int
    WarpInComplete: int

class Result(Enum):
    Victory: int
    Defeat: int
    Tie: int
    Undecided: int

class PlayerResult(Message):
    player_id: int
    result: int
    def __init__(self, player_id: int = ..., result: int = ...) -> None: ...



from __future__ import annotations

import asyncio
import sys
from contextlib import suppress
from typing import overload

from aiohttp.client_ws import ClientWebSocketResponse
from loguru import logger

# pyre-fixme[21]
from s2clientprotocol import sc2api_pb2 as sc_pb
from s2clientprotocol.query_pb2 import RequestQuery
from sc2.data import Status


class ProtocolError(Exception):
    @property
    def is_game_over_error(self) -> bool:
        return self.args[0] in ["['Game has already ended']", "['Not supported if game has already ended']"]


class ConnectionAlreadyClosedError(ProtocolError):
    pass


class Protocol:
    def __init__(self, ws: ClientWebSocketResponse) -> None:
        """
        A class for communicating with an SCII application.
        :param ws: the websocket (type: aiohttp.ClientWebSocketResponse) used to communicate with a specific SCII app
        """
        assert ws
        self._ws: ClientWebSocketResponse = ws
        # pyre-fixme[11]
        self._status: Status | None = None

    async def __request(self, request: sc_pb.Request) -> sc_pb.Response:
        logger.debug(f"Sending request: {request!r}")
        try:
            await self._ws.send_bytes(request.SerializeToString())
        except TypeError as exc:
            logger.exception("Cannot send: Connection already closed.")
            raise ConnectionAlreadyClosedError("Connection already closed.") from exc
        logger.debug("Request sent")

        response = sc_pb.Response()
        try:
            response_bytes = await self._ws.receive_bytes()
        except TypeError as exc:
            if self._status == Status.ended:
                logger.info("Cannot receive: Game has already ended.")
                raise ConnectionAlreadyClosedError("Game has already ended") from exc
            logger.error("Cannot receive: Connection already closed.")
            raise ConnectionAlreadyClosedError("Connection already closed.") from exc
        except asyncio.CancelledError:
            # If request is sent, the response must be received before reraising cancel
            try:
                await self._ws.receive_bytes()
            except asyncio.CancelledError:
                logger.critical("Requests must not be cancelled multiple times")
                sys.exit(2)
            raise

        response.ParseFromString(response_bytes)
        logger.debug("Response received")
        return response

    @overload
    async def _execute(self, create_game: sc_pb.RequestCreateGame) -> sc_pb.Response: ...
    @overload
    async def _execute(self, join_game: sc_pb.RequestJoinGame) -> sc_pb.Response: ...
    @overload
    async def _execute(self, restart_game: sc_pb.RequestRestartGame) -> sc_pb.Response: ...
    @overload
    async def _execute(self, start_replay: sc_pb.RequestStartReplay) -> sc_pb.Response: ...
    @overload
    async def _execute(self, leave_game: sc_pb.RequestLeaveGame) -> sc_pb.Response: ...
    @overload
    async def _execute(self, quick_save: sc_pb.RequestQuickSave) -> sc_pb.Response: ...
    @overload
    async def _execute(self, quick_load: sc_pb.RequestQuickLoad) -> sc_pb.Response: ...
    @overload
    async def _execute(self, quit: sc_pb.RequestQuit) -> sc_pb.Response: ...
    @overload
    async def _execute(self, game_info: sc_pb.RequestGameInfo) -> sc_pb.Response: ...
    @overload
    async def _execute(self, action: sc_pb.RequestAction) -> sc_pb.Response: ...
    @overload
    async def _execute(self, observation: sc_pb.RequestObservation) -> sc_pb.Response: ...
    @overload
    async def _execute(self, obs_action: sc_pb.RequestObserverAction) -> sc_pb.Response: ...
    @overload
    async def _execute(self, step: sc_pb.RequestStep) -> sc_pb.Response: ...
    @overload
    async def _execute(self, data: sc_pb.RequestData) -> sc_pb.Response: ...
    @overload
    async def _execute(self, query: RequestQuery) -> sc_pb.Response: ...
    @overload
    async def _execute(self, save_replay: sc_pb.RequestSaveReplay) -> sc_pb.Response: ...
    @overload
    async def _execute(self, map_command: sc_pb.RequestMapCommand) -> sc_pb.Response: ...
    @overload
    async def _execute(self, replay_info: sc_pb.RequestReplayInfo) -> sc_pb.Response: ...
    @overload
    async def _execute(self, available_maps: sc_pb.RequestAvailableMaps) -> sc_pb.Response: ...
    @overload
    async def _execute(self, save_map: sc_pb.RequestSaveMap) -> sc_pb.Response: ...
    @overload
    async def _execute(self, ping: sc_pb.RequestPing) -> sc_pb.Response: ...
    @overload
    async def _execute(self, debug: sc_pb.RequestDebug) -> sc_pb.Response: ...
    async def _execute(self, **kwargs) -> sc_pb.Response:
        assert len(kwargs) == 1, "Only one request allowed by the API"

        response = await self.__request(sc_pb.Request(**kwargs))

        new_status = Status(response.status)
        if new_status != self._status:
            logger.info(f"Client status changed to {new_status} (was {self._status})")
        self._status = new_status

        if response.error:
            logger.debug(f"Response contained an error: {response.error}")
            raise ProtocolError(f"{response.error}")

        return response

    async def ping(self):
        result = await self._execute(ping=sc_pb.RequestPing())
        return result

    async def quit(self) -> None:
        with suppress(ConnectionAlreadyClosedError, ConnectionResetError):
            await self._execute(quit=sc_pb.RequestQuit())



# pyre-ignore-all-errors[11, 16]
from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from itertools import chain

from loguru import logger

from s2clientprotocol import raw_pb2, sc2api_pb2
from sc2.constants import IS_ENEMY, IS_MINE, FakeEffectID, FakeEffectRadii
from sc2.data import Alliance, DisplayType
from sc2.ids.ability_id import AbilityId
from sc2.ids.effect_id import EffectId
from sc2.ids.upgrade_id import UpgradeId
from sc2.pixel_map import PixelMap
from sc2.position import Point2, Point3
from sc2.power_source import PsionicMatrix
from sc2.score import ScoreDetails

try:
    from sc2.dicts.generic_redirect_abilities import GENERIC_REDIRECT_ABILITIES
except ImportError:
    logger.info('Unable to import "GENERIC_REDIRECT_ABILITIES"')
    GENERIC_REDIRECT_ABILITIES = {}


class Blip:
    def __init__(self, proto: raw_pb2.Unit) -> None:
        """
        :param proto:
        """
        self._proto = proto

    @property
    def is_blip(self) -> bool:
        """Detected by sensor tower."""
        return self._proto.is_blip

    @property
    def is_snapshot(self) -> bool:
        return self._proto.display_type == DisplayType.Snapshot.value

    @property
    def is_visible(self) -> bool:
        return self._proto.display_type == DisplayType.Visible.value

    @property
    def alliance(self) -> Alliance:
        return self._proto.alliance

    @property
    def is_mine(self) -> bool:
        return self._proto.alliance == Alliance.Self.value

    @property
    def is_enemy(self) -> bool:
        return self._proto.alliance == Alliance.Enemy.value

    @property
    def position(self) -> Point2:
        """2d position of the blip."""
        return Point2.from_proto(self._proto.pos)

    @property
    def position3d(self) -> Point3:
        """3d position of the blip."""
        return Point3.from_proto(self._proto.pos)


class Common:
    ATTRIBUTES = [
        "player_id",
        "minerals",
        "vespene",
        "food_cap",
        "food_used",
        "food_army",
        "food_workers",
        "idle_worker_count",
        "army_count",
        "warp_gate_count",
        "larva_count",
    ]

    def __init__(self, proto: sc2api_pb2.PlayerCommon) -> None:
        self._proto = proto

    def __getattr__(self, attr) -> int:
        assert attr in self.ATTRIBUTES, f"'{attr}' is not a valid attribute"
        return int(getattr(self._proto, attr))


class EffectData:
    def __init__(self, proto: raw_pb2.Effect | raw_pb2.Unit, fake: bool = False) -> None:
        """
        :param proto:
        :param fake:
        """
        self._proto = proto
        self.fake = fake

    @property
    def id(self) -> EffectId | str:
        if isinstance(self._proto, raw_pb2.Unit):
            # Returns the string from constants.py, e.g. "KD8CHARGE"
            return FakeEffectID[self._proto.unit_type]
        return EffectId(self._proto.effect_id)

    @property
    def positions(self) -> set[Point2]:
        if isinstance(self._proto, raw_pb2.Unit):
            return {Point2.from_proto(self._proto.pos)}
        return {Point2.from_proto(p) for p in self._proto.pos}

    @property
    def alliance(self) -> Alliance:
        return Alliance(self._proto.alliance)

    @property
    def is_mine(self) -> bool:
        """Checks if the effect is caused by me."""
        return self._proto.alliance == IS_MINE

    @property
    def is_enemy(self) -> bool:
        """Checks if the effect is hostile."""
        return self._proto.alliance == IS_ENEMY

    @property
    def owner(self) -> int:
        return self._proto.owner

    @property
    def radius(self) -> float:
        if self.fake:
            return FakeEffectRadii[self._proto.unit_type]
        return self._proto.radius

    def __repr__(self) -> str:
        return f"{self.id} with radius {self.radius} at {self.positions}"


@dataclass
class ChatMessage:
    player_id: int
    message: str


@dataclass
class AbilityLookupTemplateClass:
    @property
    def exact_id(self) -> AbilityId:
        return AbilityId(self.ability_id)

    @property
    def generic_id(self) -> AbilityId:
        """
        See https://github.com/BurnySc2/python-sc2/blob/511c34f6b7ae51bd11e06ba91b6a9624dc04a0c0/sc2/dicts/generic_redirect_abilities.py#L13
        """
        return GENERIC_REDIRECT_ABILITIES.get(self.exact_id, self.exact_id)


@dataclass
class ActionRawUnitCommand(AbilityLookupTemplateClass):
    game_loop: int
    ability_id: int
    unit_tags: list[int]
    queue_command: bool
    target_world_space_pos: Point2 | None
    target_unit_tag: int | None = None


@dataclass
class ActionRawToggleAutocast(AbilityLookupTemplateClass):
    game_loop: int
    ability_id: int
    unit_tags: list[int]


@dataclass
class ActionRawCameraMove:
    center_world_space: Point2


@dataclass
class ActionError(AbilityLookupTemplateClass):
    ability_id: int
    unit_tag: int
    # See here for the codes of 'result': https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/error.proto#L6
    result: int


class GameState:
    def __init__(
        self,
        response_observation: sc2api_pb2.ResponseObservation,
        previous_observation: sc2api_pb2.ResponseObservation | None = None,
    ) -> None:
        """
        :param response_observation:
        :param previous_observation:
        """
        # Only filled in realtime=True in case the bot skips frames
        self.previous_observation = previous_observation
        self.response_observation = response_observation

        # https://github.com/Blizzard/s2client-proto/blob/51662231c0965eba47d5183ed0a6336d5ae6b640/s2clientprotocol/sc2api.proto#L575
        self.observation = response_observation.observation
        self.observation_raw = self.observation.raw_data
        self.player_result = response_observation.player_result
        self.common: Common = Common(self.observation.player_common)

        # Area covered by Pylons and Warpprisms
        self.psionic_matrix: PsionicMatrix = PsionicMatrix.from_proto(self.observation_raw.player.power_sources)
        # 22.4 per second on faster game speed
        self.game_loop: int = self.observation.game_loop

        # https://github.com/Blizzard/s2client-proto/blob/33f0ecf615aa06ca845ffe4739ef3133f37265a9/s2clientprotocol/score.proto#L31
        self.score: ScoreDetails = ScoreDetails(self.observation.score)
        self.abilities = self.observation.abilities  # abilities of selected units
        self.upgrades: set[UpgradeId] = {UpgradeId(upgrade) for upgrade in self.observation_raw.player.upgrade_ids}

        # self.visibility[point]: 0=Hidden, 1=Fogged, 2=Visible
        self.visibility: PixelMap = PixelMap(self.observation_raw.map_state.visibility)
        # self.creep[point]: 0=No creep, 1=creep
        self.creep: PixelMap = PixelMap(self.observation_raw.map_state.creep, in_bits=True)

        # Effects like ravager bile shot, lurker attack, everything in effect_id.py
        self.effects: set[EffectData] = {EffectData(effect) for effect in self.observation_raw.effects}
        """ Usage:
        for effect in self.state.effects:
            if effect.id == EffectId.RAVAGERCORROSIVEBILECP:
                positions = effect.positions
                # dodge the ravager biles
        """

    @cached_property
    def dead_units(self) -> set[int]:
        """A set of unit tags that died this frame"""
        _dead_units = set(self.observation_raw.event.dead_units)
        if self.previous_observation:
            return _dead_units | set(self.previous_observation.observation.raw_data.event.dead_units)
        return _dead_units

    @cached_property
    def chat(self) -> list[ChatMessage]:
        """List of chat messages sent this frame (by either player)."""
        previous_frame_chat = self.previous_observation.chat if self.previous_observation else []
        return [
            ChatMessage(message.player_id, message.message)
            for message in chain(previous_frame_chat, self.response_observation.chat)
        ]

    @cached_property
    def alerts(self) -> list[int]:
        """
        Game alerts, see https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/sc2api.proto#L683-L706
        """
        if self.previous_observation is not None:
            return list(chain(self.previous_observation.observation.alerts, self.observation.alerts))
        return self.observation.alerts

    @cached_property
    def actions(self) -> list[ActionRawUnitCommand | ActionRawToggleAutocast | ActionRawCameraMove]:
        """
        List of successful actions since last frame.
        See https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/sc2api.proto#L630-L637

        Each action is converted into Python dataclasses: ActionRawUnitCommand, ActionRawToggleAutocast, ActionRawCameraMove
        """
        previous_frame_actions = self.previous_observation.actions if self.previous_observation else []
        actions: list[ActionRawUnitCommand | ActionRawToggleAutocast | ActionRawCameraMove] = []
        for action in chain(previous_frame_actions, self.response_observation.actions):
            action_raw = action.action_raw
            game_loop = action.game_loop
            if action_raw.HasField("unit_command"):
                # Unit commands
                raw_unit_command = action_raw.unit_command
                if raw_unit_command.HasField("target_world_space_pos"):
                    # Actions that have a point as target
                    actions.append(
                        ActionRawUnitCommand(
                            game_loop,
                            raw_unit_command.ability_id,
                            raw_unit_command.unit_tags,
                            raw_unit_command.queue_command,
                            Point2.from_proto(raw_unit_command.target_world_space_pos),
                        )
                    )
                else:
                    # Actions that have a unit as target
                    actions.append(
                        ActionRawUnitCommand(
                            game_loop,
                            raw_unit_command.ability_id,
                            raw_unit_command.unit_tags,
                            raw_unit_command.queue_command,
                            None,
                            raw_unit_command.target_unit_tag,
                        )
                    )
            elif action_raw.HasField("toggle_autocast"):
                # Toggle autocast actions
                raw_toggle_autocast_action = action_raw.toggle_autocast
                actions.append(
                    ActionRawToggleAutocast(
                        game_loop,
                        raw_toggle_autocast_action.ability_id,
                        raw_toggle_autocast_action.unit_tags,
                    )
                )
            else:
                # Camera move actions
                actions.append(ActionRawCameraMove(Point2.from_proto(action.action_raw.camera_move.center_world_space)))
        return actions

    @cached_property
    def actions_unit_commands(self) -> list[ActionRawUnitCommand]:
        """
        List of successful unit actions since last frame.
        See https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/raw.proto#L185-L193
        """
        # pyre-ignore[7]
        return list(filter(lambda action: isinstance(action, ActionRawUnitCommand), self.actions))

    @cached_property
    def actions_toggle_autocast(self) -> list[ActionRawToggleAutocast]:
        """
        List of successful autocast toggle actions since last frame.
        See https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/raw.proto#L199-L202
        """
        # pyre-ignore[7]
        return list(filter(lambda action: isinstance(action, ActionRawToggleAutocast), self.actions))

    @cached_property
    def action_errors(self) -> list[ActionError]:
        """
        List of erroneous actions since last frame.
        See https://github.com/Blizzard/s2client-proto/blob/01ab351e21c786648e4c6693d4aad023a176d45c/s2clientprotocol/sc2api.proto#L648-L652
        """
        previous_frame_errors = self.previous_observation.action_errors if self.previous_observation else []
        return [
            ActionError(error.ability_id, error.unit_tag, error.result)
            for error in chain(self.response_observation.action_errors, previous_frame_errors)
        ]



"""
You can execute this test running the following command from the root python-sc2 folder:
uv run pytest test/test_pickled_ramp.py

This test/script uses the pickle files located in "python-sc2/test/pickle_data" generated by "generate_pickle_files_bot.py" file, which is a bot that starts a game on each of the maps defined in the main function.

It will load the pickle files, recreate the bot object from scratch and tests most of the bot properties and functions.
All functions that require some kind of query or interaction with the API directly will have to be tested in the "autotest_bot.py" in a live game.
"""

import time
from pathlib import Path

from loguru import logger

from sc2.game_info import Ramp
from sc2.ids.unit_typeid import UnitTypeId
from sc2.position import Point2
from sc2.unit import Unit
from sc2.units import Units
from test.test_pickled_data import MAPS, get_map_specific_bot


# From https://docs.pytest.org/en/latest/example/parametrize.html#a-quick-port-of-testscenarios
def pytest_generate_tests(metafunc):
    idlist = []
    argvalues = []
    for scenario in metafunc.cls.scenarios:
        idlist.append(scenario[0])
        items = scenario[1].items()
        argnames = [x[0] for x in items]
        argvalues.append([x[1] for x in items])
    metafunc.parametrize(argnames, argvalues, ids=idlist, scope="class")


class TestClass:
    # Load all pickle files and convert them into bot objects from raw data (game_data, game_info, game_state)
    scenarios = [(map_path.name, {"map_path": map_path}) for map_path in MAPS]

    MAPS_WITH_ODD_EXPANSION_COUNT: set[UnitTypeId] = {"Persephone AIE", "StargazersAIE", "Stasis LE"}

    def test_main_base_ramp(self, map_path: Path):
        bot = get_map_specific_bot(map_path)
        # pyre-ignore[16]
        bot.game_info.map_ramps, bot.game_info.vision_blockers = bot.game_info._find_ramps_and_vision_blockers()

        # Test if main ramp works for all spawns
        for spawn in bot.game_info.start_locations + [bot.townhalls[0].position]:
            # Remove cached precalculated ramp
            if hasattr(bot, "main_base_ramp"):
                del bot.main_base_ramp

            # Set start location as one of the opponent spawns
            bot.game_info.player_start_location = spawn

            # Find main base ramp for opponent
            ramp: Ramp = bot.main_base_ramp
            assert ramp.top_center
            assert ramp.bottom_center
            assert ramp.size
            assert ramp.points
            assert ramp.upper
            assert ramp.lower
            # Test if ramp was detected far away
            logger.info(ramp.top_center)
            distance = ramp.top_center.distance_to(bot.game_info.player_start_location)
            assert distance < 30, (
                f"Distance from spawn to main ramp was detected as {distance:.2f}, which is too far. Spawn: {spawn}, Ramp: {ramp.top_center}"
            )
            # On the map HonorgroundsLE, the main base is large and it would take a bit of effort to fix, so it returns None or empty set
            if len(ramp.upper) in {2, 5}:
                assert ramp.upper2_for_ramp_wall
                # Check if terran wall was found
                assert ramp.barracks_correct_placement
                assert ramp.barracks_in_middle
                assert ramp.depot_in_middle
                assert len(ramp.corner_depots) == 2
                # Check if protoss wall was found
                assert ramp.protoss_wall_pylon
                assert len(ramp.protoss_wall_buildings) == 2
                assert ramp.protoss_wall_warpin
            else:
                # On maps it is unable to find valid wall positions (Honorgrounds LE) it should return None, empty sets or empty lists
                assert ramp.barracks_correct_placement is None
                assert ramp.barracks_in_middle is None
                assert ramp.depot_in_middle is None
                assert ramp.corner_depots == set()
                assert ramp.protoss_wall_pylon is None
                assert ramp.protoss_wall_buildings == frozenset()
                assert ramp.protoss_wall_warpin is None

    def test_bot_ai(self, map_path: Path):
        bot = get_map_specific_bot(map_path)

        # Recalculate and time expansion locations
        t0 = time.perf_counter()
        bot._find_expansion_locations()
        t1 = time.perf_counter()
        logger.info(f"Time to calculate expansion locations: {t1 - t0} s")

        # TODO: Cache all expansion positions for a map and check if it is the same
        # BelShirVestigeLE has only 10 bases - perhaps it should be removed since it was a WOL / HOTS map
        assert len(bot.expansion_locations_list) >= 10, f"Too few expansions found: {len(bot.expansion_locations_list)}"
        # Honorgrounds LE has 24 bases
        assert len(bot.expansion_locations_list) <= 24, (
            f"Too many expansions found: {len(bot.expansion_locations_list)}"
        )
        # On N player maps, it is expected that there are N*X bases because of symmetry, at least for maps designed for 1vs1
        # Those maps in the list have an un-even expansion count
        # pyre-ignore[16]
        expect_even_expansion_count = 1 if bot.game_info.map_name in self.MAPS_WITH_ODD_EXPANSION_COUNT else 0
        assert (
            len(bot.expansion_locations_list) % (len(bot.enemy_start_locations) + 1) == expect_even_expansion_count
        ), f"{bot.expansion_locations_list}"
        # Test if bot start location is in expansion locations
        assert bot.townhalls.random.position in set(bot.expansion_locations_list), (
            f'This error might occur if you are running the tests locally using command "pytest test/", possibly because you are using an outdated cache.py version, but it should not occur when using docker and uv.\n{bot.townhalls.random.position}, {bot.expansion_locations_list}'
        )
        # Test if enemy start locations are in expansion locations
        for location in bot.enemy_start_locations:
            assert location in set(bot.expansion_locations_list), f"{location}, {bot.expansion_locations_list}"
        # Each expansion is supposed to have at least one geysir and 6-12 minerals
        # pyre-ignore[16]
        for expansion, resource_positions in bot.expansion_locations_dict.items():
            assert isinstance(expansion, Point2)
            assert isinstance(resource_positions, Units)
            if resource_positions:
                assert isinstance(resource_positions[0], Unit)
            # 2000 Atmospheres has bases with just 4 minerals patches and a rich geysir
            # Neon violet has bases with just 6 resources. I think that was the back corner base with 4 minerals and 2 vespene
            # Odyssey has bases with 10 mineral patches and 2 geysirs
            # Blood boil returns 21?
            assert 5 <= len(resource_positions) <= 12, (
                f"{len(resource_positions)} resource fields in one base on map {bot.game_info.map_name}"
            )

        assert bot.owned_expansions == {bot.townhalls.first.position: bot.townhalls.first}



# pyre-ignore-all-errors[6, 11, 16, 21, 29]
from __future__ import annotations

import asyncio
import json
import platform
import signal
import sys
from contextlib import suppress
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Any

import mpyq
import portpicker
from aiohttp import ClientSession, ClientWebSocketResponse
from loguru import logger

from s2clientprotocol import sc2api_pb2 as sc_pb
from sc2.bot_ai import BotAI
from sc2.client import Client
from sc2.controller import Controller
from sc2.data import CreateGameError, Result, Status
from sc2.game_state import GameState
from sc2.maps import Map
from sc2.observer_ai import ObserverAI
from sc2.player import AbstractPlayer, Bot, BotProcess, Human
from sc2.portconfig import Portconfig
from sc2.protocol import ConnectionAlreadyClosedError, ProtocolError
from sc2.proxy import Proxy
from sc2.sc2process import KillSwitch, SC2Process

# Set the global logging level
logger.remove()
logger.add(sys.stdout, level="INFO")


@dataclass
class GameMatch:
    """Dataclass for hosting a match of SC2.
    This contains all of the needed information for RequestCreateGame.
    :param sc2_config: dicts of arguments to unpack into sc2process's construction, one per player
        second sc2_config will be ignored if only one sc2_instance is spawned
        e.g. sc2_args=[{"fullscreen": True}, {}]: only player 1's sc2instance will be fullscreen
    :param game_time_limit: The time (in seconds) until a match is artificially declared a Tie
    """

    map_sc2: Map
    players: list[AbstractPlayer]
    realtime: bool = False
    random_seed: int | None = None
    disable_fog: bool | None = None
    sc2_config: list[dict] | None = None
    game_time_limit: int | None = None

    def __post_init__(self) -> None:
        # avoid players sharing names
        if len(self.players) > 1 and self.players[0].name is not None and self.players[0].name == self.players[1].name:
            self.players[1].name += "2"

        if self.sc2_config is not None:
            if isinstance(self.sc2_config, dict):
                self.sc2_config = [self.sc2_config]
            if len(self.sc2_config) == 0:
                self.sc2_config = [{}]
            while len(self.sc2_config) < len(self.players):
                self.sc2_config += self.sc2_config
            self.sc2_config = self.sc2_config[: len(self.players)]

    @property
    def needed_sc2_count(self) -> int:
        return sum(player.needs_sc2 for player in self.players)

    @property
    def host_game_kwargs(self) -> dict[str, Any]:
        return {
            "map_settings": self.map_sc2,
            "players": self.players,
            "realtime": self.realtime,
            "random_seed": self.random_seed,
            "disable_fog": self.disable_fog,
        }

    def __repr__(self) -> str:
        p1 = self.players[0]
        p1 = p1.name if p1.name else p1
        p2 = self.players[1]
        p2 = p2.name if p2.name else p2
        return f"Map: {self.map_sc2.name}, {p1} vs {p2}, realtime={self.realtime}, seed={self.random_seed}"


async def _play_game_human(client, player_id, realtime, game_time_limit):
    while True:
        state = await client.observation()
        if client._game_result:
            return client._game_result[player_id]

        if game_time_limit and state.observation.observation.game_loop / 22.4 > game_time_limit:
            logger.info(state.observation.game_loop, state.observation.game_loop / 22.4)
            return Result.Tie

        if not realtime:
            await client.step()


async def _play_game_ai(
    client: Client, player_id: int, ai: BotAI, realtime: bool, game_time_limit: int | None
) -> Result:
    gs: GameState | None = None

    async def initialize_first_step() -> Result | None:
        nonlocal gs
        ai._initialize_variables()

        game_data = await client.get_game_data()
        game_info = await client.get_game_info()
        ping_response = await client.ping()

        # This game_data will become self.game_data in botAI
        ai._prepare_start(
            client, player_id, game_info, game_data, realtime=realtime, base_build=ping_response.ping.base_build
        )
        state = await client.observation()
        # check game result every time we get the observation
        if client._game_result:
            await ai.on_end(client._game_result[player_id])
            return client._game_result[player_id]
        gs = GameState(state.observation)
        proto_game_info = await client._execute(game_info=sc_pb.RequestGameInfo())
        try:
            ai._prepare_step(gs, proto_game_info)
            await ai.on_before_start()
            ai._prepare_first_step()
            await ai.on_start()
        # TODO Catching too general exception Exception (broad-except)

        except Exception as e:
            logger.exception(f"Caught unknown exception in AI on_start: {e}")
            logger.error("Resigning due to previous error")
            await ai.on_end(Result.Defeat)
            return Result.Defeat

    result = await initialize_first_step()
    if result is not None:
        return result

    async def run_bot_iteration(iteration: int):
        nonlocal gs
        logger.debug(f"Running AI step, it={iteration} {gs.game_loop / 22.4:.2f}s")
        # Issue event like unit created or unit destroyed
        await ai.issue_events()
        # In on_step various errors can occur - log properly
        try:
            await ai.on_step(iteration)
        except (AttributeError,) as e:
            logger.exception(f"Caught exception: {e}")
            raise
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")
            raise
        await ai._after_step()
        logger.debug("Running AI step: done")

    # Only used in realtime=True
    previous_state_observation = None
    for iteration in range(10**10):
        if realtime and gs:
            # On realtime=True, might get an error here: sc2.protocol.ProtocolError: ['Not in a game']
            with suppress(ProtocolError):
                requested_step = gs.game_loop + client.game_step
                state = await client.observation(requested_step)
                # If the bot took too long in the previous observation, request another observation one frame after
                if state.observation.observation.game_loop > requested_step:
                    logger.debug("Skipped a step in realtime=True")
                    previous_state_observation = state.observation
                    state = await client.observation(state.observation.observation.game_loop + 1)
        else:
            state = await client.observation()

        # check game result every time we get the observation
        if client._game_result:
            await ai.on_end(client._game_result[player_id])
            return client._game_result[player_id]
        gs = GameState(state.observation, previous_state_observation)
        previous_state_observation = None
        logger.debug(f"Score: {gs.score.score}")

        if game_time_limit and gs.game_loop / 22.4 > game_time_limit:
            await ai.on_end(Result.Tie)
            return Result.Tie
        proto_game_info = await client._execute(game_info=sc_pb.RequestGameInfo())
        ai._prepare_step(gs, proto_game_info)

        await run_bot_iteration(iteration)  # Main bot loop

        if not realtime:
            if not client.in_game:  # Client left (resigned) the game
                await ai.on_end(client._game_result[player_id])
                return client._game_result[player_id]

            # TODO: In bot vs bot, if the other bot ends the game, this bot gets stuck in requesting an observation when using main.py:run_multiple_games
            await client.step()
    return Result.Undecided


async def _play_game(
    player: AbstractPlayer,
    client: Client,
    realtime: bool,
    portconfig: Portconfig,
    game_time_limit: int | None = None,
    rgb_render_config: dict[str, Any] | None = None,
) -> Result:
    assert isinstance(realtime, bool), repr(realtime)

    player_id = await client.join_game(
        player.name, player.race, portconfig=portconfig, rgb_render_config=rgb_render_config
    )
    logger.info(f"Player {player_id} - {player.name if player.name else str(player)}")

    if isinstance(player, Human):
        result = await _play_game_human(client, player_id, realtime, game_time_limit)
    else:
        result = await _play_game_ai(client, player_id, player.ai, realtime, game_time_limit)

    logger.info(
        f"Result for player {player_id} - {player.name if player.name else str(player)}: "
        f"{result._name_ if isinstance(result, Result) else result}"
    )

    return result


async def _play_replay(client: Client, ai, realtime: bool = False, player_id: int = 0):
    ai._initialize_variables()

    game_data = await client.get_game_data()
    game_info = await client.get_game_info()
    ping_response = await client.ping()

    client.game_step = 1
    # This game_data will become self._game_data in botAI
    ai._prepare_start(
        client, player_id, game_info, game_data, realtime=realtime, base_build=ping_response.ping.base_build
    )
    state = await client.observation()
    # Check game result every time we get the observation
    if client._game_result:
        await ai.on_end(client._game_result[player_id])
        return client._game_result[player_id]
    gs = GameState(state.observation)
    proto_game_info = await client._execute(game_info=sc_pb.RequestGameInfo())
    ai._prepare_step(gs, proto_game_info)
    ai._prepare_first_step()
    try:
        await ai.on_start()
    # TODO Catching too general exception Exception (broad-except)

    except Exception as e:
        logger.exception(f"Caught unknown exception in AI replay on_start: {e}")
        await ai.on_end(Result.Defeat)
        return Result.Defeat

    iteration = 0
    while True:
        if iteration != 0:
            if realtime:
                # TODO: check what happens if a bot takes too long to respond, so that the requested
                #  game_loop might already be in the past
                state = await client.observation(gs.game_loop + client.game_step)
            else:
                state = await client.observation()
            # check game result every time we get the observation
            if client._game_result:
                try:
                    await ai.on_end(client._game_result[player_id])
                except TypeError:
                    return client._game_result[player_id]
                return client._game_result[player_id]
            gs = GameState(state.observation)
            logger.debug(f"Score: {gs.score.score}")

            proto_game_info = await client._execute(game_info=sc_pb.RequestGameInfo())
            ai._prepare_step(gs, proto_game_info)

        logger.debug(f"Running AI step, it={iteration} {gs.game_loop * 0.725 * (1 / 16):.2f}s")

        try:
            # Issue event like unit created or unit destroyed
            await ai.issue_events()
            await ai.on_step(iteration)
            await ai._after_step()

        # TODO Catching too general exception Exception (broad-except)
        except Exception as e:
            if isinstance(e, ProtocolError) and e.is_game_over_error:
                if realtime:
                    return None
                await ai.on_end(Result.Victory)
                return None
            # NOTE: this message is caught by pytest suite
            logger.exception("AI step threw an error")  # DO NOT EDIT!
            logger.error(f"Error: {e}")
            logger.error("Resigning due to previous error")
            try:
                await ai.on_end(Result.Defeat)
            except TypeError:
                return Result.Defeat
            return Result.Defeat

        logger.debug("Running AI step: done")

        if not realtime and not client.in_game:  # Client left (resigned) the game
            await ai.on_end(Result.Victory)
            return Result.Victory

        await client.step()  # unindent one line to work in realtime

        iteration += 1


async def _setup_host_game(
    server: Controller, map_settings, players, realtime, random_seed=None, disable_fog=None, save_replay_as=None
):
    r = await server.create_game(map_settings, players, realtime, random_seed, disable_fog)
    if r.create_game.HasField("error"):
        err = f"Could not create game: {CreateGameError(r.create_game.error)}"
        if r.create_game.HasField("error_details"):
            err += f": {r.create_game.error_details}"
        logger.critical(err)
        raise RuntimeError(err)

    return Client(server._ws, save_replay_as)


async def _host_game(
    map_settings: Map,
    players: list[AbstractPlayer],
    realtime: bool = False,
    portconfig: Portconfig | None = None,
    save_replay_as: str | None = None,
    game_time_limit: int | None = None,
    rgb_render_config: dict[str, Any] | None = None,
    random_seed: int | None = None,
    sc2_version: str | None = None,
    disable_fog: bool = False,
):
    assert players, "Can't create a game without players"

    assert any((isinstance(p, (Human, Bot))) for p in players)

    async with SC2Process(
        fullscreen=players[0].fullscreen, render=rgb_render_config is not None, sc2_version=sc2_version
    ) as server:
        await server.ping()

        client = await _setup_host_game(
            server, map_settings, players, realtime, random_seed, disable_fog, save_replay_as
        )
        # Bot can decide if it wants to launch with 'raw_affects_selection=True'
        if not isinstance(players[0], Human) and getattr(players[0].ai, "raw_affects_selection", None) is not None:
            client.raw_affects_selection = players[0].ai.raw_affects_selection

        result = await _play_game(players[0], client, realtime, portconfig, game_time_limit, rgb_render_config)
        if client.save_replay_path is not None:
            await client.save_replay(client.save_replay_path)
        try:
            await client.leave()
        except ConnectionAlreadyClosedError:
            logger.error("Connection was closed before the game ended")
        await client.quit()

        return result


async def _host_game_aiter(
    map_settings,
    players,
    realtime,
    portconfig=None,
    save_replay_as=None,
    game_time_limit=None,
):
    assert players, "Can't create a game without players"

    assert any(isinstance(p, (Human, Bot)) for p in players)

    async with SC2Process() as server:
        while True:
            await server.ping()

            client = await _setup_host_game(server, map_settings, players, realtime)
            if not isinstance(players[0], Human) and getattr(players[0].ai, "raw_affects_selection", None) is not None:
                client.raw_affects_selection = players[0].ai.raw_affects_selection

            try:
                result = await _play_game(players[0], client, realtime, portconfig, game_time_limit)

                if save_replay_as is not None:
                    await client.save_replay(save_replay_as)
                await client.leave()
            except ConnectionAlreadyClosedError:
                logger.error("Connection was closed before the game ended")
                return

            new_players = yield result
            if new_players is not None:
                players = new_players


def _host_game_iter(*args, **kwargs):
    game = _host_game_aiter(*args, **kwargs)
    new_playerconfig = None
    while True:
        new_playerconfig = yield asyncio.get_event_loop().run_until_complete(game.asend(new_playerconfig))


async def _join_game(
    players: list[AbstractPlayer],
    realtime: bool = False,
    portconfig: Portconfig | None = None,
    save_replay_as: str | None = None,
    game_time_limit: int | None = None,
    sc2_version: str | None = None,
):
    async with SC2Process(fullscreen=players[1].fullscreen, sc2_version=sc2_version) as server:
        await server.ping()

        client = Client(server._ws)
        # Bot can decide if it wants to launch with 'raw_affects_selection=True'
        if isinstance(players[1], Bot) and getattr(players[1].ai, "raw_affects_selection", None) is not None:
            client.raw_affects_selection = players[1].ai.raw_affects_selection

        result = await _play_game(players[1], client, realtime, portconfig, game_time_limit)
        if save_replay_as is not None:
            await client.save_replay(save_replay_as)
        try:
            await client.leave()
        except ConnectionAlreadyClosedError:
            logger.error("Connection was closed before the game ended")
        await client.quit()

        return result


async def _setup_replay(server, replay_path, realtime, observed_id):
    await server.start_replay(replay_path, realtime, observed_id)
    return Client(server._ws)


async def _host_replay(
    replay_path, ai: ObserverAI, realtime: bool, _portconfig: Portconfig, base_build, data_version, observed_id
):
    async with SC2Process(fullscreen=False, base_build=base_build, data_hash=data_version) as server:
        client = await _setup_replay(server, replay_path, realtime, observed_id)
        result = await _play_replay(client, ai, realtime)
        return result


def get_replay_version(replay_path: str | Path) -> tuple[str, str]:
    with Path(replay_path).open("rb") as f:
        replay_data = f.read()
        replay_io = BytesIO()
        replay_io.write(replay_data)
        replay_io.seek(0)
        archive = mpyq.MPQArchive(replay_io).extract()
        metadata = json.loads(archive[b"replay.gamemetadata.json"].decode("utf-8"))
        return metadata["BaseBuild"], metadata["DataVersion"]


# TODO Deprecate run_game function in favor of run_multiple_games
def run_game(
    map_settings: Map,
    players: list[AbstractPlayer],
    realtime: bool,
    portconfig: Portconfig | None = None,
    save_replay_as: str | None = None,
    game_time_limit: int | None = None,
    rgb_render_config: dict[str, Any] | None = None,
    random_seed: int | None = None,
    sc2_version: str | None = None,
    disable_fog: bool = False,
) -> Result | list[Result | None]:
    """
    Returns a single Result enum if the game was against the built-in computer.
    Returns a list of two Result enums if the game was "Human vs Bot" or "Bot vs Bot".
    """
    if sum(isinstance(p, (Human, Bot)) for p in players) > 1:
        portconfig = Portconfig()

        async def run_host_and_join():
            return await asyncio.gather(
                _host_game(
                    map_settings,
                    players,
                    realtime=realtime,
                    portconfig=portconfig,
                    save_replay_as=save_replay_as,
                    game_time_limit=game_time_limit,
                    rgb_render_config=rgb_render_config,
                    random_seed=random_seed,
                    sc2_version=sc2_version,
                    disable_fog=disable_fog,
                ),
                _join_game(
                    players,
                    realtime=realtime,
                    portconfig=portconfig,
                    save_replay_as=save_replay_as,
                    game_time_limit=game_time_limit,
                    sc2_version=sc2_version,
                ),
                return_exceptions=True,
            )

        result: list[Result] = asyncio.run(run_host_and_join())
        assert isinstance(result, list)
        assert all(isinstance(r, Result) for r in result)
    else:
        result: Result = asyncio.run(
            _host_game(
                map_settings,
                players,
                realtime=realtime,
                portconfig=portconfig,
                save_replay_as=save_replay_as,
                game_time_limit=game_time_limit,
                rgb_render_config=rgb_render_config,
                random_seed=random_seed,
                sc2_version=sc2_version,
                disable_fog=disable_fog,
            )
        )
        assert isinstance(result, Result)
    return result


def run_replay(ai: ObserverAI, replay_path: Path | str, realtime: bool = False, observed_id: int = 0):
    portconfig = Portconfig()
    assert Path(replay_path).is_file(), f"Replay does not exist at the given path: {replay_path}"
    assert Path(replay_path).is_absolute(), (
        f'Replay path has to be an absolute path, e.g. "C:/replays/my_replay.SC2Replay" but given path was "{replay_path}"'
    )
    base_build, data_version = get_replay_version(replay_path)
    result = asyncio.get_event_loop().run_until_complete(
        _host_replay(replay_path, ai, realtime, portconfig, base_build, data_version, observed_id)
    )
    return result


async def play_from_websocket(
    ws_connection: str | ClientWebSocketResponse,
    player: AbstractPlayer,
    realtime: bool = False,
    portconfig: Portconfig | None = None,
    save_replay_as: str | None = None,
    game_time_limit: int | None = None,
    should_close: bool = True,
):
    """Use this to play when the match is handled externally e.g. for bot ladder games.
    Portconfig MUST be specified if not playing vs Computer.
    :param ws_connection: either a string("ws://{address}:{port}/sc2api") or a ClientWebSocketResponse object
    :param should_close: closes the connection if True. Use False if something else will reuse the connection

    e.g. ladder usage: play_from_websocket("ws://127.0.0.1:5162/sc2api", MyBot, False, portconfig=my_PC)
    """
    session = None
    try:
        if isinstance(ws_connection, str):
            session = ClientSession()
            ws_connection = await session.ws_connect(ws_connection, timeout=120)
            should_close = True
        client = Client(ws_connection)
        result = await _play_game(player, client, realtime, portconfig, game_time_limit=game_time_limit)
        if save_replay_as is not None:
            await client.save_replay(save_replay_as)
    except ConnectionAlreadyClosedError:
        logger.error("Connection was closed before the game ended")
        return None
    finally:
        if should_close:
            await ws_connection.close()
            if session:
                await session.close()

    return result


async def run_match(controllers: list[Controller], match: GameMatch, close_ws: bool = True):
    await _setup_host_game(controllers[0], **match.host_game_kwargs)

    # Setup portconfig beforehand, so all players use the same ports
    startport = None
    portconfig = None
    if match.needed_sc2_count > 1:
        if any(isinstance(player, BotProcess) for player in match.players):
            portconfig = Portconfig.contiguous_ports()
            # Most ladder bots generate their server and client ports as [s+2, s+3], [s+4, s+5]
            startport = portconfig.server[0] - 2
        else:
            portconfig = Portconfig()

    proxies = []
    coros = []
    players_that_need_sc2 = filter(lambda lambda_player: lambda_player.needs_sc2, match.players)
    for i, player in enumerate(players_that_need_sc2):
        if isinstance(player, BotProcess):
            pport = portpicker.pick_unused_port()
            p = Proxy(controllers[i], player, pport, match.game_time_limit, match.realtime)
            proxies.append(p)
            coros.append(p.play_with_proxy(startport))
        else:
            coros.append(
                play_from_websocket(
                    controllers[i]._ws,
                    player,
                    match.realtime,
                    portconfig,
                    should_close=close_ws,
                    game_time_limit=match.game_time_limit,
                )
            )

    async_results = await asyncio.gather(*coros, return_exceptions=True)

    if not isinstance(async_results, list):
        async_results = [async_results]
    for i, a in enumerate(async_results):
        if isinstance(a, Exception):
            logger.error(f"Exception[{a}] thrown by {[p for p in match.players if p.needs_sc2][i]}")

    return process_results(match.players, async_results)


def process_results(players: list[AbstractPlayer], async_results: list[Result]) -> dict[AbstractPlayer, Result]:
    opp_res = {Result.Victory: Result.Defeat, Result.Defeat: Result.Victory, Result.Tie: Result.Tie}
    result: dict[AbstractPlayer, Result] = {}
    i = 0
    for player in players:
        if player.needs_sc2:
            if sum(r == Result.Victory for r in async_results) <= 1:
                result[player] = async_results[i]
            else:
                result[player] = Result.Undecided
            i += 1
        else:  # computer
            other_result = async_results[0]
            result[player] = None
            if other_result in opp_res:
                result[player] = opp_res[other_result]

    return result


async def maintain_SCII_count(count: int, controllers: list[Controller], proc_args: list[dict] | None = None) -> None:
    """Modifies the given list of controllers to reflect the desired amount of SCII processes"""
    # kill unhealthy ones.
    if controllers:
        to_remove = []
        alive = await asyncio.wait_for(
            asyncio.gather(*(c.ping() for c in controllers if not c._ws.closed), return_exceptions=True), timeout=20
        )
        i = 0  # for alive
        for controller in controllers:
            if controller._ws.closed:
                if not controller._process._session.closed:
                    await controller._process._session.close()
                to_remove.append(controller)
            else:
                if not isinstance(alive[i], sc_pb.Response):
                    try:
                        await controller._process._close_connection()
                    finally:
                        to_remove.append(controller)
                i += 1
        for c in to_remove:
            c._process._clean(verbose=False)
            if c._process in KillSwitch._to_kill:
                KillSwitch._to_kill.remove(c._process)
            controllers.remove(c)

    # spawn more
    if len(controllers) < count:
        needed = count - len(controllers)
        if proc_args:
            index = len(controllers) % len(proc_args)
        else:
            proc_args = [{} for _ in range(needed)]
            index = 0
        extra = [SC2Process(**proc_args[(index + _) % len(proc_args)]) for _ in range(needed)]
        logger.info(f"Creating {needed} more SC2 Processes")
        for _ in range(3):
            if platform.system() == "Linux":
                # Works on linux: start one client after the other

                new_controllers = [await asyncio.wait_for(sc.__aenter__(), timeout=50) for sc in extra]
            else:
                # Doesnt seem to work on linux: starting 2 clients nearly at the same time
                new_controllers = await asyncio.wait_for(
                    asyncio.gather(*[sc.__aenter__() for sc in extra], return_exceptions=True),
                    timeout=50,
                )

            controllers.extend(c for c in new_controllers if isinstance(c, Controller))
            if len(controllers) == count:
                await asyncio.wait_for(asyncio.gather(*(c.ping() for c in controllers)), timeout=20)
                break
            extra = [
                extra[i] for i, result in enumerate(new_controllers) if not isinstance(new_controllers, Controller)
            ]
        else:
            logger.critical("Could not launch sufficient SC2")
            raise RuntimeError

    # kill excess
    while len(controllers) > count:
        proc = controllers.pop()
        proc = proc._process
        logger.info(f"Removing SCII listening to {proc._port}")
        await proc._close_connection()
        proc._clean(verbose=False)
        if proc in KillSwitch._to_kill:
            KillSwitch._to_kill.remove(proc)


def run_multiple_games(matches: list[GameMatch]):
    return asyncio.get_event_loop().run_until_complete(a_run_multiple_games(matches))


# TODO Catching too general exception Exception (broad-except)


async def a_run_multiple_games(matches: list[GameMatch]) -> list[dict[AbstractPlayer, Result]]:
    """Run multiple matches.
    Non-python bots are supported.
    When playing bot vs bot, this is less likely to fatally crash than repeating run_game()
    """
    if not matches:
        return []

    results = []
    controllers = []
    for m in matches:
        result = None
        dont_restart = m.needed_sc2_count == 2
        try:
            await maintain_SCII_count(m.needed_sc2_count, controllers, m.sc2_config)
            result = await run_match(controllers, m, close_ws=dont_restart)
        except SystemExit as e:
            logger.info(f"Game exit'ed as {e} during match {m}")
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")
            logger.info(f"Exception {e} thrown in match {m}")
        finally:
            if dont_restart:  # Keeping them alive after a non-computer match can cause crashes
                await maintain_SCII_count(0, controllers, m.sc2_config)
            results.append(result)
    KillSwitch.kill_all()
    return results


# TODO Catching too general exception Exception (broad-except)


async def a_run_multiple_games_nokill(matches: list[GameMatch]) -> list[dict[AbstractPlayer, Result]]:
    """Run multiple matches while reusing SCII processes.
    Prone to crashes and stalls
    """
    # FIXME: check whether crashes between bot-vs-bot are avoidable or not
    if not matches:
        return []

    # Start the matches
    results = []
    controllers: list[Controller] = []
    for m in matches:
        logger.info(f"Starting match {1 + len(results)} / {len(matches)}: {m}")
        result = None
        try:
            await maintain_SCII_count(m.needed_sc2_count, controllers, m.sc2_config)
            result = await run_match(controllers, m, close_ws=False)
        except SystemExit as e:
            logger.critical(f"Game sys.exit'ed as {e} during match {m}")
        except Exception as e:
            logger.exception(f"Caught unknown exception: {e}")
            logger.info(f"Exception {e} thrown in match {m}")
        finally:
            for c in controllers:
                try:
                    await c.ping()
                    if c._status != Status.launched:
                        await c._execute(leave_game=sc_pb.RequestLeaveGame())
                except Exception as e:
                    logger.exception(f"Caught unknown exception: {e}")
                    if not (isinstance(e, ProtocolError) and e.is_game_over_error):
                        logger.info(f"controller {c.__dict__} threw {e}")

            results.append(result)

    # Fire the killswitch manually, instead of letting the winning player fire it.
    await asyncio.wait_for(asyncio.gather(*(c._process._close_connection() for c in controllers)), timeout=50)
    KillSwitch.kill_all()
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    return results



from enum import Enum

class ActionResult(Enum):
    Success: int
    NotSupported: int
    Error: int
    CantQueueThatOrder: int
    Retry: int
    Cooldown: int
    QueueIsFull: int
    RallyQueueIsFull: int
    NotEnoughMinerals: int
    NotEnoughVespene: int
    NotEnoughTerrazine: int
    NotEnoughCustom: int
    NotEnoughFood: int
    FoodUsageImpossible: int
    NotEnoughLife: int
    NotEnoughShields: int
    NotEnoughEnergy: int
    LifeSuppressed: int
    ShieldsSuppressed: int
    EnergySuppressed: int
    NotEnoughCharges: int
    CantAddMoreCharges: int
    TooMuchMinerals: int
    TooMuchVespene: int
    TooMuchTerrazine: int
    TooMuchCustom: int
    TooMuchFood: int
    TooMuchLife: int
    TooMuchShields: int
    TooMuchEnergy: int
    MustTargetUnitWithLife: int
    MustTargetUnitWithShields: int
    MustTargetUnitWithEnergy: int
    CantTrade: int
    CantSpend: int
    CantTargetThatUnit: int
    CouldntAllocateUnit: int
    UnitCantMove: int
    TransportIsHoldingPosition: int
    BuildTechRequirementsNotMet: int
    CantFindPlacementLocation: int
    CantBuildOnThat: int
    CantBuildTooCloseToDropOff: int
    CantBuildLocationInvalid: int
    CantSeeBuildLocation: int
    CantBuildTooCloseToCreepSource: int
    CantBuildTooCloseToResources: int
    CantBuildTooFarFromWater: int
    CantBuildTooFarFromCreepSource: int
    CantBuildTooFarFromBuildPowerSource: int
    CantBuildOnDenseTerrain: int
    CantTrainTooFarFromTrainPowerSource: int
    CantLandLocationInvalid: int
    CantSeeLandLocation: int
    CantLandTooCloseToCreepSource: int
    CantLandTooCloseToResources: int
    CantLandTooFarFromWater: int
    CantLandTooFarFromCreepSource: int
    CantLandTooFarFromBuildPowerSource: int
    CantLandTooFarFromTrainPowerSource: int
    CantLandOnDenseTerrain: int
    AddOnTooFarFromBuilding: int
    MustBuildRefineryFirst: int
    BuildingIsUnderConstruction: int
    CantFindDropOff: int
    CantLoadOtherPlayersUnits: int
    NotEnoughRoomToLoadUnit: int
    CantUnloadUnitsThere: int
    CantWarpInUnitsThere: int
    CantLoadImmobileUnits: int
    CantRechargeImmobileUnits: int
    CantRechargeUnderConstructionUnits: int
    CantLoadThatUnit: int
    NoCargoToUnload: int
    LoadAllNoTargetsFound: int
    NotWhileOccupied: int
    CantAttackWithoutAmmo: int
    CantHoldAnyMoreAmmo: int
    TechRequirementsNotMet: int
    MustLockdownUnitFirst: int
    MustTargetUnit: int
    MustTargetInventory: int
    MustTargetVisibleUnit: int
    MustTargetVisibleLocation: int
    MustTargetWalkableLocation: int
    MustTargetPawnableUnit: int
    YouCantControlThatUnit: int
    YouCantIssueCommandsToThatUnit: int
    MustTargetResources: int
    RequiresHealTarget: int
    RequiresRepairTarget: int
    NoItemsToDrop: int
    CantHoldAnyMoreItems: int
    CantHoldThat: int
    TargetHasNoInventory: int
    CantDropThisItem: int
    CantMoveThisItem: int
    CantPawnThisUnit: int
    MustTargetCaster: int
    CantTargetCaster: int
    MustTargetOuter: int
    CantTargetOuter: int
    MustTargetYourOwnUnits: int
    CantTargetYourOwnUnits: int
    MustTargetFriendlyUnits: int
    CantTargetFriendlyUnits: int
    MustTargetNeutralUnits: int
    CantTargetNeutralUnits: int
    MustTargetEnemyUnits: int
    CantTargetEnemyUnits: int
    MustTargetAirUnits: int
    CantTargetAirUnits: int
    MustTargetGroundUnits: int
    CantTargetGroundUnits: int
    MustTargetStructures: int
    CantTargetStructures: int
    MustTargetLightUnits: int
    CantTargetLightUnits: int
    MustTargetArmoredUnits: int
    CantTargetArmoredUnits: int
    MustTargetBiologicalUnits: int
    CantTargetBiologicalUnits: int
    MustTargetHeroicUnits: int
    CantTargetHeroicUnits: int
    MustTargetRoboticUnits: int
    CantTargetRoboticUnits: int
    MustTargetMechanicalUnits: int
    CantTargetMechanicalUnits: int
    MustTargetPsionicUnits: int
    CantTargetPsionicUnits: int
    MustTargetMassiveUnits: int
    CantTargetMassiveUnits: int
    MustTargetMissile: int
    CantTargetMissile: int
    MustTargetWorkerUnits: int
    CantTargetWorkerUnits: int
    MustTargetEnergyCapableUnits: int
    CantTargetEnergyCapableUnits: int
    MustTargetShieldCapableUnits: int
    CantTargetShieldCapableUnits: int
    MustTargetFlyers: int
    CantTargetFlyers: int
    MustTargetBuriedUnits: int
    CantTargetBuriedUnits: int
    MustTargetCloakedUnits: int
    CantTargetCloakedUnits: int
    MustTargetUnitsInAStasisField: int
    CantTargetUnitsInAStasisField: int
    MustTargetUnderConstructionUnits: int
    CantTargetUnderConstructionUnits: int
    MustTargetDeadUnits: int
    CantTargetDeadUnits: int
    MustTargetRevivableUnits: int
    CantTargetRevivableUnits: int
    MustTargetHiddenUnits: int
    CantTargetHiddenUnits: int
    CantRechargeOtherPlayersUnits: int
    MustTargetHallucinations: int
    CantTargetHallucinations: int
    MustTargetInvulnerableUnits: int
    CantTargetInvulnerableUnits: int
    MustTargetDetectedUnits: int
    CantTargetDetectedUnits: int
    CantTargetUnitWithEnergy: int
    CantTargetUnitWithShields: int
    MustTargetUncommandableUnits: int
    CantTargetUncommandableUnits: int
    MustTargetPreventDefeatUnits: int
    CantTargetPreventDefeatUnits: int
    MustTargetPreventRevealUnits: int
    CantTargetPreventRevealUnits: int
    MustTargetPassiveUnits: int
    CantTargetPassiveUnits: int
    MustTargetStunnedUnits: int
    CantTargetStunnedUnits: int
    MustTargetSummonedUnits: int
    CantTargetSummonedUnits: int
    MustTargetUser1: int
    CantTargetUser1: int
    MustTargetUnstoppableUnits: int
    CantTargetUnstoppableUnits: int
    MustTargetResistantUnits: int
    CantTargetResistantUnits: int
    MustTargetDazedUnits: int
    CantTargetDazedUnits: int
    CantLockdown: int
    CantMindControl: int
    MustTargetDestructibles: int
    CantTargetDestructibles: int
    MustTargetItems: int
    CantTargetItems: int
    NoCalldownAvailable: int
    WaypointListFull: int
    MustTargetRace: int
    CantTargetRace: int
    MustTargetSimilarUnits: int
    CantTargetSimilarUnits: int
    CantFindEnoughTargets: int
    AlreadySpawningLarva: int
    CantTargetExhaustedResources: int
    CantUseMinimap: int
    CantUseInfoPanel: int
    OrderQueueIsFull: int
    CantHarvestThatResource: int
    HarvestersNotRequired: int
    AlreadyTargeted: int
    CantAttackWeaponsDisabled: int
    CouldntReachTarget: int
    TargetIsOutOfRange: int
    TargetIsTooClose: int
    TargetIsOutOfArc: int
    CantFindTeleportLocation: int
    InvalidItemClass: int
    CantFindCancelOrder: int



import random
from math import atan2, cos, pi, sin, sqrt

from sc2.position import EPSILON, Point2

P0 = Point2((0, 0))
P1 = Point2((1, 1))
P2 = Point2((3, 1))
P3 = Point2((3, 3))


def rad_diff(a, b):
    r1 = abs(a - b)
    r2 = abs(b - a)
    r3 = min(r1, r2)
    r4 = abs(2 * pi - r3)
    return min(r3, r4)


def test_test_rad_diff():
    assert rad_diff(0, 0) == 0
    assert rad_diff(0, 1) == 1
    assert rad_diff(0, pi) == pi
    assert rad_diff(pi, pi) == 0
    assert rad_diff(2 * pi, 0) == 0
    assert rad_diff(2 * pi + 1, 0) == 1
    assert rad_diff(2 * pi + 1, 1) == 0
    assert rad_diff(pi, -pi) == 0


def test_distance():
    assert P0.distance_to(P1) == sqrt(2)
    assert P1.distance_to(P2) == 2
    assert P0.distance_to(P2) == sqrt(10)


def test_towards():
    assert P0.towards(P1, 1) == Point2((sqrt(2) / 2, sqrt(2) / 2))


def test_random_on_distance():
    random.seed(1)

    def get_points(source, distance, n=1000):
        return {source.random_on_distance(distance) for _ in range(n)}

    def verify_distances(source, distance, n=1000):
        for p in get_points(source, distance, n):
            assert abs(source.distance_to(p) - distance) < 0.000001

    def verify_angles(source, distance, n=1000):
        angles_rad = {atan2(p.y - source.y, p.x - source.x) for p in get_points(source, distance, n)}

        quadrants = {(cos(a) < 0, sin(a) < 0) for a in angles_rad}
        assert len(quadrants) == 4

    verify_distances(P0, 1e2)
    verify_distances(P1, 1e3)
    verify_distances(P2, 1e4)

    verify_angles(P0, 1e2)
    verify_angles(P1, 1e3)
    verify_angles(P2, 1e4)


def test_towards_random_angle():
    random.seed(1)

    def random_points(n=1000):
        def rs():
            return 1 - random.random() * 2

        return {Point2((rs() * 1000, rs() * 1000)) for _ in range(n)}

    def verify(source, target, max_difference=(pi / 4), n=1000):
        d = 1 + random.random() * 100
        points = {source.towards_with_random_angle(target, distance=d, max_difference=max_difference) for _ in range(n)}

        dx, dy = target.x - source.x, target.y - source.y
        src_angle = atan2(dy, dx)

        for p in points:
            angle = atan2(p.y - source.y, p.x - source.x)
            assert rad_diff(src_angle, angle) <= max_difference

            assert abs(source.distance_to(p) - d) <= EPSILON

    verify(P0, P1)
    verify(P1, P2)
    verify(P1, P3)
    verify(P2, P3)

    verify(P1, P0)
    verify(P2, P1)
    verify(P3, P1)
    verify(P3, P2)

    ps = random_points(n=50)
    for p1 in ps:
        for p2 in ps:
            if p1 == p2:
                continue
            verify(p1, p2, n=10)



from pathlib import Path


def is_submodule(path):
    if path.is_file():
        return path.suffix == ".py" and path.stem != "__init__"
    if path.is_dir():
        return (path / "__init__.py").exists()
    return False


__all__ = [p.stem for p in Path(__file__).parent.iterdir() if is_submodule(p)]
